package googccwrap;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

public class ClassLoaderTool {

    private final static String COMPILER_JAR = "/compiler-20100917.jar";

    private static File jarFile;

    private static ClosureCompiler compiler;

    /**
     * Extracts the jar file for the compiler from the classpath into a temporary jar file that can is deleted on vm
     * exit.
     * 
     * @return The absolute path of the extracted jar file.
     */
    public static synchronized File extractCompilerJar() {

        if (jarFile != null) {
            return jarFile;
        }

        InputStream is = ClassLoaderTool.class.getResourceAsStream(COMPILER_JAR);

        if (is == null) {
            throw new IllegalStateException("Missing classpath resource: " + COMPILER_JAR);
        }

        try {

            File tmp = File.createTempFile("compiler", "jar");
            tmp.deleteOnExit();

            FileOutputStream fos = new FileOutputStream(tmp);

            Util.spool(is, fos, true);

            return tmp;

        } catch (IOException e) {
            throw new IllegalArgumentException("Failed to extract jar", e);
        }

    }

    public static synchronized ClosureCompiler getCompiler() {

        if (compiler != null) {
            return compiler;
        }

        File jarFile = extractCompilerJar();

        URL url;
        try {
            url = jarFile.toURI().toURL();
        } catch (MalformedURLException e) {
            throw new RuntimeException(e);
        }

        GoogleClosureClassLoader cl = new GoogleClosureClassLoader(Thread.currentThread().getContextClassLoader(), url);

        try {

            Class<?> runnerClass = cl.loadClass("com.google.javascript.jscomp.CommandLineRunner");

            final Constructor<?> runnerConst = runnerClass.getDeclaredConstructor(String[].class, PrintStream.class,
                    PrintStream.class);
            runnerConst.setAccessible(true);

            final Method doRun = runnerClass.getSuperclass().getDeclaredMethod("doRun");
            doRun.setAccessible(true);

            compiler = new ClosureCompiler() {

                @Override
                public InputStream compile(String[] args, InputStream is) throws CompilationFailedException {

                    PrintStream out = null;
                    PrintStream err = null;
                    File tmpFile = null;
                    int status = 0;

                    try {

                        tmpFile = File.createTempFile("googccwrap-", "js");
                        tmpFile.deleteOnExit();
                        OutputStream tmpOut = new BufferedOutputStream(new FileOutputStream(tmpFile));
                        Util.spool(is, tmpOut, true);
                        tmpOut.close();

                        List<String> argList = new LinkedList<String>();
                        int skip = 0;
                        boolean hasWarnLevel = false;
                        for (String arg : args) {
                            while (skip-- > 0) {
                                continue;
                            }
                            if (arg.equalsIgnoreCase("--js_output_file")) {
                                skip = 1;
                                continue;
                            } else if (arg.equalsIgnoreCase("--js")) {
                                skip = 1;
                                continue;
                            } else if (arg.equalsIgnoreCase("--warning_level")) {
                                hasWarnLevel = true;
                            }
                            argList.add(arg);
                        }
                        if (!hasWarnLevel) {
                            argList.add("--warning_level");
                            argList.add("QUIET");
                        }
                        argList.add("--js");
                        argList.add(tmpFile.getAbsolutePath());
                        args = argList.toArray(new String[argList.size()]);

                        ByteArrayOutputStream baosOut = new ByteArrayOutputStream();
                        out = new PrintStream(baosOut);

                        ByteArrayOutputStream baosErr = new ByteArrayOutputStream();
                        err = new PrintStream(baosErr);

                        Object inst = runnerConst.newInstance(args, out, err);

                        status = (Integer) doRun.invoke(inst);

                        if (status != 0) {
                            throw new CompilationFailedException(status, new String(baosErr.toByteArray(), "utf-8"));
                        }

                        return new ByteArrayInputStream(baosOut.toByteArray());

                    } catch (CompilationFailedException ce) {
                        throw ce;
                    } catch (Exception e) {
                        throw new RuntimeException(e);
                    } finally {
                        try {
                            // leave files that failed to compile (at least until vm exits)
                            if (tmpFile != null && status == 0)
                                tmpFile.delete();
                            if (out != null)
                                out.close();
                            if (err != null)
                                err.close();
                        } catch (Exception e) {
                            // at least we tried
                        }
                    }

                }

            };

        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        return compiler;

    }

    private static class GoogleClosureClassLoader extends URLClassLoader {

        GoogleClosureClassLoader(ClassLoader parent, URL... url) {
            super(url, parent);
        }

        @Override
        protected synchronized Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
            // First, check if the class has already been loaded
            Class<?> c = findLoadedClass(name);
            if (c == null) {
                if (isLoadedFromParent(name)) {
                    try {
                        if (getParent() != null) {
                            c = getParent().loadClass(name);
                        }
                    } catch (ClassNotFoundException e) {
                        // ClassNotFoundException thrown if class not found
                        // from the non-null parent class loader
                    }
                }
                if (c == null) {
                    // If still not found, then invoke findClass in order
                    // to find the class.
                    c = findClass(name);
                }
            }
            if (resolve) {
                resolveClass(c);
            }
            return c;
        }

        // these classes are loaded from the jar in the local class loader, the rest
        // is delegated up.
        private final static Pattern LOCAL_CLASSES =
                Pattern.compile("^(com\\.google|org\\.kohsuke|javax\\.annotation|org\\.mozilla|" +
                        "rhino_ast|externs.zip|build-data.properties).*");

        private boolean isLoadedFromParent(String name) {

            return !LOCAL_CLASSES.matcher(name).matches();

        }

    }

}
